
package com.yahoo.shopping;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import com.yahoo.prelude.ChainedSearcher;
import com.yahoo.prelude.Hit;
import com.yahoo.prelude.Query;
import com.yahoo.prelude.Result;
import com.yahoo.prelude.query.AndItem;
import com.yahoo.prelude.query.CompositeItem;
import com.yahoo.prelude.query.IntItem;
import com.yahoo.prelude.query.Item;
import com.yahoo.prelude.query.PhraseItem;
import com.yahoo.prelude.query.TermItem;
import com.yahoo.prelude.query.WordItem;

/**
 * The XSearcher plug-in provides functionality to focus the search: 
 * - Allows for context queries, i.e. queries that include terms that describe the desired context, with no specific knowledge 
 * 	 about the query language or document structure needed.
 * - Keyword and context terms in a query is treated differently, using the context terms only for focusing the search.
 * - Provides a generic and flexible framework for hybrid search.
 * @author Anne Siri Korsen & Trond ivind Eriksen.
 *
 */
public class XSearcher extends ChainedSearcher{
	
	private HashMap<String, String> xmlDocumentTree;
	private HashMap<String, String> ontology;
	private TreeMap<String, String> invertedXmlDocumentTree;
	private HashMap<String, String> firstNamesMap;
	private HashMap<String, String> surNamesMap;
	private HashMap<String, String> depMap;
	
	
	/**
	 * The constructor initiates the data structures. Three main data structures are used: 
	 * - A hash map storing mappings from context terms to their Dewey numbers. 
	 * - A hash map that acts as an ontology, storing mappings from synonyms or plural forms to the terms' base form.
	 * - A sorted hash map storing mappings from each Dewey number to its XML label.
	 * The first two mentioned hash maps are used during query analysis, in order to recognise context terms and express the 
	 * desired search context by means of the Dewey number of its root node. The sorted hash map is used to traverse the part 
	 * of the XML tree to be searched in, retrieving subresults, and merging these when appropriate. In addition, two hash maps 
	 * are built storing lists of English first names and surnames.
	 */
	public XSearcher() {
		
		xmlDocumentTree = new HashMap<String, String>();
		xmlDocumentTree.put("documents", "0");
		xmlDocumentTree.put("dvd", "0.0");
		xmlDocumentTree.put("actor", "0.0.0");
		xmlDocumentTree.put("director", "0.0.1");
		xmlDocumentTree.put("music", "0.1");
		xmlDocumentTree.put("artist", "0.1.0");
		xmlDocumentTree.put("book", "0.2");
		xmlDocumentTree.put("author", "0.2.0");
		
		ontology = new HashMap<String, String>();
		ontology.put("dvd", "dvd");
		ontology.put("dvds", "dvd");
		ontology.put("movie", "dvd");
		ontology.put("movies", "dvd");
		ontology.put("video", "dvd");
		ontology.put("videos", "dvd");
		ontology.put("actor", "actor");
		ontology.put("actors", "actor");
		ontology.put("director", "director");
		ontology.put("directors", "director");
		ontology.put("music", "music");
		ontology.put("album", "music");		
		ontology.put("albums", "music");
		ontology.put("artist", "artist");
		ontology.put("artists", "artist");
		ontology.put("book", "book");
		ontology.put("books", "book");
		ontology.put("author", "author");
		ontology.put("authors", "author");
		
		invertedXmlDocumentTree = new TreeMap<String, String>();
		invertedXmlDocumentTree.put("0", "documents");
		invertedXmlDocumentTree.put("0.0", "dvd");
		invertedXmlDocumentTree.put("0.0.0", "actor");
		invertedXmlDocumentTree.put("0.0.1", "director");
		invertedXmlDocumentTree.put("0.0.2", "title");
		invertedXmlDocumentTree.put("0.0.3", "desc");
		invertedXmlDocumentTree.put("0.1", "music");
		invertedXmlDocumentTree.put("0.1.0", "artist");
		invertedXmlDocumentTree.put("0.1.1", "title");
		invertedXmlDocumentTree.put("0.1.2", "desc");
		invertedXmlDocumentTree.put("0.2", "book");
		invertedXmlDocumentTree.put("0.2.0", "author");
		invertedXmlDocumentTree.put("0.2.1", "title");
		invertedXmlDocumentTree.put("0.2.2", "desc");
		
		depMap = new HashMap<String, String>();
		depMap.put("0.0.0", "65");
		depMap.put("0.0.1", "65");
		depMap.put("0.0.2", "65");
		depMap.put("0.0.3", "65");
		depMap.put("0.1.0", "66");
		depMap.put("0.1.1", "66");
		depMap.put("0.1.2", "66");
		depMap.put("0.2.0", "56");
		depMap.put("0.2.1", "56");
		depMap.put("0.2.2", "56");
		
		firstNamesMap = new HashMap<String,String>();
		surNamesMap = new HashMap<String,String>();
		buildNamesMap("/home/y/lib/qrs/plugins/com/yahoo/shopping/data/surnames_eng.txt");
		buildNamesMap("/home/y/lib/qrs/plugins/com/yahoo/shopping/data/firstnames_eng.txt");
		System.out.println("firstNames length: "+firstNamesMap.size());
		System.out.println("surNames length: "+surNamesMap.size());
		
	}
	
	/**
	 * This method is called each time a search is performed. It determines the values of the input parameters, and initiates 
	 * query analysis and searching by subcalls to analyseQuery() and doXmlSearch(). 
	 * @param query Query as it is forwarded from the previous Searcher in the chain.
	 * @param offset The first hit number to retrieve.
	 * @param hits The number of hits to retrieve in each subresult.
	 * @return The final result returned to the previous Searcher in the chain.
	 */
	public Result doSearch(Query query,int offset,int hits) {	
		
		if(query.getBooleanProperty("xsearch")) {
		
			boolean properName = false;
			AggregationFunction aggrFunc = new Avg();
			int k = 10;		
			
			if(query.getBooleanProperty("propername")) { properName = true;}
			
			if(query.getStringProperty("aggrfunc") != null) {
				if(query.getStringProperty("aggrfunc").equals("avg")) { aggrFunc = new Avg(); } 
				else if (query.getStringProperty("aggrfunc").equals("max")) { aggrFunc = new Max(); }
			}
			
			if(query.getStringProperty("top-k") != null) { k = Integer.parseInt(query.getStringProperty("top-k")); }
			
			System.out.println("ProperName: "+properName);
			System.out.println("AggrFunc: "+aggrFunc.toString());
			System.out.println("Top-k: "+k);
			
			String[] searchContext = analyseQuery(query, properName);
			
			System.out.println("searchcontext: "+searchContext[0]+", "+searchContext[1]);
			
			Result result = doXmlSearch(query, searchContext[0], Integer.parseInt(searchContext[1]), offset, hits, k, aggrFunc);
			
			return result;
			
			
		} else {
			
			Result result = super.doSearch(query, offset, hits);
			return result;
			
		}
	}
	
	
	/**
	 * The method reads a file, and stores the value of each line in a HashMap.
	 * The values are either first names or surnames. 
	 * @param fileName The file name that is to be opened and read from. 
	 */
	private void buildNamesMap(String fileName){
		try {
			String line = null; 
			BufferedReader br = new BufferedReader(new FileReader(fileName)); 
			while ((line = br.readLine()) != null) { 
				if(fileName.equalsIgnoreCase("/home/y/lib/qrs/plugins/com/yahoo/shopping/data/surnames_eng.txt"))
					surNamesMap.put(line, line);
				if(fileName.equalsIgnoreCase("/home/y/lib/qrs/plugins/com/yahoo/shopping/data/firstnames_eng.txt"))
					firstNamesMap.put(line, line); 
			} 
			System.out.println("Finished building map.");
			br.close(); 
			br = null; 
			
		} catch (FileNotFoundException e) {
			System.out.println("File '" + fileName + "' does not exist. "); 
		} catch (IOException ioex) {
			ioex.printStackTrace();
		}		
	}
	
	
	/**
	 * The method recognises context terms and proper names.
	 * @param query The raw query to be traversed.
	 * @param properName Whether proper name recognition should be utilised.
	 * @return The desired search context represented by the Dewey number of its root node. 
	 */
	private String[] analyseQuery(Query query, boolean properName) {
		
		Item rawRoot = query.getRawRoot();
		boolean foundLabel = false;
		
		String label = "";
		
		/** Checking for labels **/
		if (rawRoot instanceof CompositeItem) {
			CompositeItem compositeQuery = (CompositeItem)rawRoot;
			int numQueryTerms = compositeQuery.getItemCount();
			
			/** Checking if the last query term is a label **/
			Item lastItem = compositeQuery.getItem(numQueryTerms-1);
			if(lastItem instanceof WordItem) {
				
				if( xmlDocumentTree.containsKey(ontology.get(((WordItem) lastItem).getWord().trim())) ){
					foundLabel = true;
					label = ontology.get(((WordItem) lastItem).getWord().trim());
				}
				
				if(!foundLabel) {
					Item firstItem = compositeQuery.getItem(0);
					if(firstItem instanceof WordItem) {
						if( xmlDocumentTree.containsKey(ontology.get(((WordItem)firstItem).getWord().trim())) ){
							foundLabel = true;
							label = ontology.get(((WordItem)firstItem).getWord().trim());
						}
					}
				}
			}
		}
		
		
		//checking for proper names
		int queryTermsAreAllNames = 1;
		
		if (rawRoot instanceof CompositeItem) {
			CompositeItem compositeItem = (CompositeItem)rawRoot;
			Iterator queryIt = compositeItem.getItemIterator();
			while(queryIt.hasNext()) {
				Item subItem = (Item)queryIt.next();
				if(subItem instanceof WordItem) {
					WordItem wordItem = (WordItem)subItem;
					if( (!(surNamesMap.containsKey(wordItem.getWord()) || firstNamesMap.containsKey(wordItem.getWord()))) && 
							!xmlDocumentTree.containsKey(ontology.get(((WordItem)wordItem).getWord().trim())) ){
						queryTermsAreAllNames = 0;
						break;
					}
				} 
			}
		}
		else if(rawRoot instanceof WordItem) {
			WordItem wordItem = (WordItem)rawRoot;
			if( !(surNamesMap.containsKey(wordItem.getWord()) || firstNamesMap.containsKey(wordItem.getWord())) ){
				queryTermsAreAllNames = 0;
			}
		}
		
		//build searchcontext
		String[] searchContext = new String[2];
		if(foundLabel){
			if(properName & queryTermsAreAllNames == 0) { 
				//proper name experiment is turned on - if search for a non proper name in artist, author, actor or director, 
				//steer search to root node
				if(label.equals("artist") || label.equals("author") || label.equals("actor") || label.equals("director")) {
					searchContext[0] = xmlDocumentTree.get("documents");
				}
			}
			else {
				searchContext[0] = xmlDocumentTree.get(label);
			}
		}
		else { //no label in query string
			searchContext[0] = xmlDocumentTree.get("documents");
		}
		searchContext[1] = ""+queryTermsAreAllNames;
		
		return searchContext;
	}
	
	/**
	 * Traverses the XML structure from the root context node found in analyseQuery(), performs a search by calling doXSearch() 
	 * for each leaf node, and merges the subresults by calling mergeResults().
	 * @param query The raw query. 
	 * @param contextLabel The Dewey number of the desired context node.
	 * @param properNameFound Whether all query keywords (except the context term) are proper names.
	 * @param offset The first hit number to retrieve.
	 * @param hits The number of hits to retrieve in each subresult.
	 * @param k The number of hits to return in the merged result.
	 * @param aggrFunc Which aggregation function that should be used. 
	 * @return The final result.
	 */
	private Result doXmlSearch(Query query, String contextLabel, int properNameFound, int offset, int hits, int k, AggregationFunction aggrFunc) {
		
		ArrayList<Result> results = new ArrayList<Result>();
		ArrayList<Result> subResults = new ArrayList<Result>();
		Stack<String> prefixStack = new Stack<String>();
		Result result;
		
		String nextSibling = contextLabel.substring(0, contextLabel.length()-1)+(Integer.parseInt(contextLabel.substring(contextLabel.length()-1, contextLabel.length()))+1);
		SortedMap<String, String> subTree = invertedXmlDocumentTree.subMap(contextLabel+".0", nextSibling);
		if(subTree.size() > 0) {
			
			//initalize
			prefixStack.push(contextLabel);
			
			Iterator nodeIt = subTree.keySet().iterator();	
			while(nodeIt.hasNext()) {
				Query queryClone = (Query) query.clone(); //TODO: Shallow copy "bug"
				
				String currentLabel = (String)nodeIt.next();
				String nextSubSibling = currentLabel.substring(0, currentLabel.length()-1)+(Integer.parseInt(currentLabel.substring(currentLabel.length()-1, currentLabel.length()))+1);
				SortedMap<String, String> subSubTree = invertedXmlDocumentTree.subMap(currentLabel+".0", nextSubSibling);
				if(subSubTree.size() > 0) {				
					if(currentLabel.startsWith(prefixStack.peek())) { //new subtree, push on stack
						prefixStack.push(currentLabel);
					}
					else { //new subtree, pop - push
						prefixStack.pop(); 
						Result subResult = mergeResults(queryClone, subResults, k, aggrFunc);
						results.add(subResult);
						subResults.clear();
						prefixStack.push(currentLabel);
					}
				}
				else { //leafnode
					if(currentLabel.startsWith(prefixStack.peek())) { 
						Result subResult = doXSearch(contextLabel, properNameFound, offset, hits, subResults, queryClone, currentLabel);
						subResults.add(subResult);
					}
				}
			}
			
			//finalize
			Result subResult = mergeResults(query, subResults, k, aggrFunc);
			results.add(subResult);
			result = mergeResults(query, results, k, aggrFunc);
		}
		else { //leafnode
			Result subResult = doXSearch(contextLabel, properNameFound, offset, hits, subResults, query, contextLabel);
			results.add(subResult);
			result = mergeResults(query, results, k, aggrFunc);
		}
		
		return result;
		
	}
	
	/**
	 * This method is a workaround, caused by inadequate labelling of parts of the document corpus.
	 * The Shopping vertical uses six types of document schemes. Three of these; book, music, and video, are labelled in an 
	 * appropriate way. For example, the music document type contains separate fields for artist and song. The rest of the document 
	 * types; paidmerchant, freemerchant, and fastupdate, have an inappropriate labelling scheme, only providing generic fields 
	 * as txt and int. Book, music, and video only constitute about 8% of the total number of documents. In order to have the 
	 * possibility to use the entire document base as a test collection, we implemented a workaround that changed a search from 
	 * actor:keywords to desc:keywords AND desc:actor for these document types. This was also done for the fields author, actor, and 
	 * director. Additionally, a department filter was added to focus the search towards the desired category. In the case of the keywords 
	 * being proper names, the keywords were rewritten as a phrase.
	 * @param contextLabel The Dewey number of the desired context node.
	 * @param properNameFound Whether all query keywords (except the context term) are proper names.
	 * @param offset The first hit number to retrieve.
	 * @param hits The number of hits to retrieve in each subresult. 
	 * @param query The raw query.
	 * @param currentLabel The Dewey number of the current node to be searched in.
	 * @return The subresult. 
	 */
	private Result doXSearch(String contextLabel, int properNameFound, int offset, int hits, ArrayList<Result> subResults, Query query, String currentLabel) {
		
		if(invertedXmlDocumentTree.get(currentLabel).equals("author") || invertedXmlDocumentTree.get(currentLabel).equals("artist") ||
				invertedXmlDocumentTree.get(currentLabel).equals("actor") || invertedXmlDocumentTree.get(currentLabel).equals("director")) {
			
			if(properNameFound == 1) { 
				Item item = query.getRoot();
				if(item instanceof CompositeItem) {
					CompositeItem composite = (CompositeItem)item;
					PhraseItem phraseItem = new PhraseItem();
					Iterator compositeIt = composite.getItemIterator();
					while(compositeIt.hasNext()) {
						Item subItem = (Item)compositeIt.next();
						if(subItem instanceof WordItem) {
							WordItem wordItem = (WordItem)subItem;
							if(wordItem.getWord().equals(depMap.get(currentLabel))) {} 
							else if(ontology.get(wordItem.getWord()) != null && 
									ontology.get(wordItem.getWord()).equals(invertedXmlDocumentTree.get(contextLabel))) {}
							else {
								phraseItem.addItem(wordItem);
							}
						}
					}
					composite = new AndItem();
					composite.addItem(phraseItem);
					query.setRoot(composite);
				} 
				else if(item instanceof WordItem) {
					WordItem wordItem = (WordItem)item;
					AndItem composite = new AndItem();
					composite.addItem(wordItem);
					query.setRoot(composite);
				}
				
				//search in book, music, video
				Result subResult = doSuperSearch(query, offset, hits, currentLabel, contextLabel, true);
				
				//search in paidmerchant, freemerchant, fastupdate	
				((CompositeItem)query.getRoot()).addItem(new WordItem(invertedXmlDocumentTree.get(currentLabel)));
				query.setSearchList("paidmerchant,freemerchant,fastupdate");
				
				String newLabel;
				if(invertedXmlDocumentTree.get(currentLabel).equals("author")) {
					newLabel = "0.2.2";
				}
				else if(invertedXmlDocumentTree.get(currentLabel).equals("artist")) {
					newLabel = "0.1.2";							
				}
				else {
					newLabel = "0.0.3";
				}
				
				//merge hits
				Result subResult_hack = doSuperSearch(query, offset, hits, newLabel, contextLabel, false);
				subResult.mergeWith(subResult_hack);
				Iterator subResult_hackIt = subResult_hack.getHitIterator();
				while(subResult_hackIt.hasNext()) {
					subResult.addHit((Hit)subResult_hackIt.next());
				}
				subResult.sortHits();
				return subResult;
			}
			else {
				Result subResult = doSuperSearch(query, offset, hits, currentLabel, contextLabel, true);
				return subResult;
			}
		}
		else { //title and description
			Result subResult = doSuperSearch(query, offset, hits, currentLabel, contextLabel, true);
			return subResult;
		}
	}
	
	/**
	 * Modifies the query by calling modifyQuery(), and sends it to the next Searcher in the chain.
	 * @param query The hacked query.
	 * @param offset The first hit number to retrieve.
	 * @param hits The number of hits to retrieve in each subresult.
	 * @param currentLabel The Dewey number of the current node to be searched in.
	 * @param contextLabel The Dewey number of the desired context node.
	 * @param removeContextWord Whether the context term should be removed from the query.
	 * @return The retrieved subresult.
	 */
	
	private Result doSuperSearch(Query query, int offset, int hits, String currentLabel, String contextLabel, boolean removeContextWord) {
		
		query.setRoot(modifyQuery((Item)query.getRoot(), currentLabel, contextLabel, removeContextWord));
		Result result = super.doSearch(query, offset, hits);
		return result;
		
	}
	
	/**
	 * The method traverses the query tree, and performs the following modifications: 
	 * - Removal of a keyword recognised as a context term.
	 * - Addition of the context term as index labels, e.g. music:madonna.
	 * - Addition of a department filter, e.g. department=books. This is a part of the workaround performed in doXSearch(). 
	 * @param item The hacked query root.
	 * @param currentLabel The Dewey number of the current node to be searched in.
	 * @param contextLabel The Dewey number of the desired context node.
	 * @param removeContextWord Whether the context term should be removed from the query.
	 * @return The modified query root.
	 */
	private Item modifyQuery(Item item, String currentLabel, String contextLabel, boolean removeContextWord) {
		
		AndItem andItem = null;
		if(item instanceof AndItem) {
			andItem = (AndItem)item;
			int[] removeItemIndices = {-1, -1};
			for(Iterator i=andItem.getItemIterator(); i.hasNext(); ) {
				Item subItem = (Item)i.next();
				if(subItem instanceof WordItem) {
					WordItem wordItem = (WordItem)subItem;
					if(ontology.get(wordItem.getWord()) != null && 
							ontology.get(wordItem.getWord()).equals(invertedXmlDocumentTree.get(contextLabel)) &&
							removeContextWord) {
						removeItemIndices[0] = andItem.getItemIndex(wordItem); 
					} 
					else if(wordItem.getWord().equals(depMap.get(currentLabel))) { //because of query shallow copy
						removeItemIndices[1] = andItem.getItemIndex(wordItem);
					} 
					else {
						wordItem.setIndexName(invertedXmlDocumentTree.get(currentLabel));
					}
				}
				else if(subItem instanceof IntItem) {
					IntItem intItem = (IntItem)subItem;
					intItem.setIndexName(invertedXmlDocumentTree.get(currentLabel));
				}
				else if(subItem instanceof PhraseItem) {
					PhraseItem phraseItem = (PhraseItem)subItem;
					phraseItem.setIndexName(invertedXmlDocumentTree.get(currentLabel));
				}
			}
			if(removeItemIndices[0] != -1) { andItem.removeItem(removeItemIndices[0]); }
			if(removeItemIndices[1] != -1) { andItem.removeItem(removeItemIndices[1]); }
			WordItem depItem = new WordItem(depMap.get(currentLabel), "did");
			andItem.addItem(depItem);
		} 
		else if(item instanceof WordItem) {
			WordItem wordItem = (WordItem)item;
			wordItem.setIndexName(invertedXmlDocumentTree.get(currentLabel));
			WordItem depItem = new WordItem(depMap.get(currentLabel), "did");
			andItem = new AndItem();
			andItem.addItem(wordItem);
			andItem.addItem(depItem);
		}
		else if(item instanceof IntItem) {
			IntItem intItem = (IntItem)item;
			intItem.setIndexName(invertedXmlDocumentTree.get(currentLabel));
			WordItem depItem = new WordItem(depMap.get(currentLabel), "did");
			andItem = new AndItem();
			andItem.addItem(intItem);
			andItem.addItem(depItem);
		}
		return andItem;
		
	}
	
	/**
	 * Merges the subresults by performing the No Random Access algorithm. The top-k hits are sorted according to their lower bound 
	 * values. NraHit objects are kept in a hash map, sorted according to NraHitOrderer. The method uses the AggregationFunction as 
	 * a common interface to the specific aggregation functions.
	 * @param query The original query.
	 * @param subResults The subresults to be merged. 
	 * @param k The number of hits to return in the merged result.
	 * @param aggrFunc Which aggregation function that should be used. 
	 * @return The merged result.
	 */
	private Result mergeResults(Query query, ArrayList<Result> subResults, int k, AggregationFunction aggrFunc) {
		
		if(subResults.size() > 1) {
			
			HashMap<Hit, Integer[]> foundHits = new HashMap<Hit, Integer[]>();
			TreeSet<NraHit> sortedHits = new TreeSet<NraHit>(new NraHitOrderer());
			Integer[] bottomScores = new Integer[subResults.size()];
			for(int i = 0; i < bottomScores.length; i++) {
				bottomScores[i] = 0;
			}
			
			int biggestResult = 0;
			for (int r = 0; r < subResults.size(); r++) {
				if(subResults.get(r).getHitCount() > biggestResult) { 
					biggestResult = subResults.get(r).getHitCount();
				}
			}
			
			int currentBottomLowerBound = 0;
			int currentBottomUpperBound = 0;
			
			for(int hitIt = 0; hitIt < biggestResult; hitIt++) { 
				currentBottomLowerBound = 0;
				currentBottomUpperBound = 0;
				
				for(int resultIt = 0; resultIt < subResults.size(); resultIt++) { 
					
					if(hitIt < subResults.get(resultIt).getHitCount()) { 
						
						Hit hit = subResults.get(resultIt).getHit(hitIt);
						int relevancy = Integer.parseInt(hit.getProperty("relevancy"));
						String uri = hit.getProperty("uri");
						String title = hit.getProperty("title");
						
						//maintain bottom values...
						bottomScores[resultIt] = relevancy;
						
						if(uri != null) {
							Integer[] hitScores;
							if(foundHits.containsKey(hit))	{
								hitScores = foundHits.get(hit);
								hitScores[resultIt] = relevancy;
								foundHits.put(hit, hitScores);
							} else {
								hitScores = new Integer[subResults.size()];
								for(int i = 0; i < hitScores.length; i++) {
									hitScores[i] = -1;
								}
								hitScores[resultIt] = relevancy;
								foundHits.put(hit, hitScores);
							}
						}
						
					}
				}	
				
				Iterator<Hit> foundHitsIt = foundHits.keySet().iterator();
				sortedHits.clear();
				
				while(foundHitsIt.hasNext()) {
					
					Hit hit = foundHitsIt.next();
					
					//for every item discovered so far, compute lower and upper bounds...
					Integer[] hitScores = foundHits.get(hit);
					int lowerBound = aggrFunc.aggregate(hitScores);
					int upperBound = aggrFunc.aggregate(hitScores, bottomScores);
					
					sortedHits.add(new NraHit(lowerBound, upperBound, hit));
					
				}
				
				if(sortedHits.size() >= k) {
					currentBottomLowerBound = ((NraHit)sortedHits.toArray()[k-1]).getLowerBound();
					currentBottomUpperBound = ((NraHit)sortedHits.toArray()[k-1]).getUpperBound();
				} else {
					currentBottomLowerBound = ((NraHit)sortedHits.toArray()[sortedHits.size()-1]).getLowerBound();
					currentBottomUpperBound = ((NraHit)sortedHits.toArray()[sortedHits.size()-1]).getUpperBound();
				}
				
				int upperBoundNext = aggrFunc.aggregate(bottomScores);
				if(upperBoundNext < currentBottomLowerBound) { //stopping rule
					break;
				}
				
			}
			
			Result result = new Result(query);
			//merge meta information
			for (int r = 0; r < subResults.size(); r++) {
				result.mergeWith(subResults.get(r));
			}
			
			Iterator<NraHit> sortedHitsIt = sortedHits.iterator();
			int kI = 0;
			while(sortedHitsIt.hasNext() && kI < k) {
				NraHit nraHit = sortedHitsIt.next();
				Hit hit = nraHit.getHit();
				int lowerBound = nraHit.getLowerBound();
				int upperBound = nraHit.getUpperBound();
				hit.setRelevancy(lowerBound);
				result.addHit(hit);
				kI++;
			}
			result.setTotalHitCount(result.getHitCount());
			return result;
		} //only one result
		else {
			Result result = subResults.get(0);
			int nrOfHits = result.getHitCount();
			for(int hitI = k; hitI < nrOfHits; hitI++) {
				result.removeHit(k);
			}
			result.setTotalHitCount(result.getHitCount());
			return result;
		}
		
	}
	
	/**
	 * Sorts NraHit objects, firstby lower bound, then by upper bound. 
	 * @author Anne Siri Korsen & Trond ivind Eriksen.
	 *
	 */
	private static class NraHitOrderer implements Comparator<NraHit> {
		
		/** Called by Collections to carry out the sort */
		public int compare(NraHit nraHit1, NraHit nraHit2) {
			
			if(nraHit1.getLowerBound() > nraHit2.getLowerBound()) {
				return -1;
			} else if (nraHit1.getLowerBound() == nraHit2.getLowerBound()) {
				if(nraHit1.getUpperBound() > nraHit2.getUpperBound()) {
					return -1;
				} else if(nraHit1.getUpperBound() == nraHit2.getUpperBound()) {
					return 1;
				} else {
					return 1;
				}
			} else {
				return 1;
			}
		}
		
	}
	
	
}
